/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file
/// \brief Construct a strategy for a beta distribution
///

#ifndef GRAPH_MODEL_DISTRIBUTION_BETA_BETA_DISTRIBUTION_HPP
#define GRAPH_MODEL_DISTRIBUTION_BETA_BETA_DISTRIBUTION_HPP 1

#include <boost/hana/map.hpp>
#include <graph_model/distribution/beta/beta_strategy.hpp>
#include <graph_model/distribution/beta/tags.hpp>
#include <graph_model/distribution/parameter/alpha.hpp>
#include <graph_model/distribution/parameter/beta.hpp>
#include <graph_model/distribution/parameter/make_parameter_default.hpp>
#include <graph_model/distribution/parameter/named_parameter.hpp>
#include <graph_model/distribution/parameter/parameter_pack.hpp>
#include <graph_model/distribution/parameter/precision.hpp>
#include <graph_model/distribution/parameter/probability.hpp>
#include <graph_model/distribution/parameter/variate.hpp>

namespace graph_model
{

  namespace distribution
  {

    namespace beta
    {

      namespace detail
      {

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   >
	auto beta_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>);

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Key, typename Value
		   , typename ... Parameters
		   >
	auto beta_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>,
	 parameter::named_parameter<Key,Value>,
	 Parameters&& ...);

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Value
		   , typename ... Parameters
		   >
	auto beta_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>,
	 Value&&,
	 Parameters&& ...);

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged >
	auto beta_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>);

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Key, typename Value
		   , typename ... Parameters
		   >
	auto beta_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>,
	 parameter::named_parameter<Key,Value>,
	 Parameters&& ...);

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Value
		   , typename ... Parameters
		   >
	auto beta_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>,
	 Value&&,
	 Parameters&& ...);

	template < typename Tag >
	struct retag
	{
	  template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		     , typename Value
		     , typename ... Parameters
		   >
	  static auto apply
	  (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	   parameter::named_parameter<Tag,Value> parameter,
	   Parameters&& ... parameters)
	  {
	    return beta_distribution
	      (pack.add(parameter),std::forward<Parameters>(parameters)...);
	  }
	};

	template <>
	struct retag < parameter::tag::alpha >
	{
	  template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		     , typename Value
		     , typename ... Parameters
		   >
	  static auto apply
	  (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	   parameter::named_parameter<parameter::tag::alpha,Value> parameter,
	   Parameters&& ... parameters)
	  {
	    return beta_distribution<tag::alpha_beta>
	      (pack.add(parameter),std::forward<Parameters>(parameters)...);
	  }
	};

	template <>
	struct retag < parameter::tag::beta >
	{
	  template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		     , typename Value
		     , typename ... Parameters
		   >
	  static auto apply
	  (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	   parameter::named_parameter<parameter::tag::beta,Value> parameter,
	   Parameters&& ... parameters)
	  {
	    return beta_distribution<tag::alpha_beta>
	      (pack.add(parameter),std::forward<Parameters>(parameters)...);
	  }
	};

	template <>
	struct retag < parameter::tag::probability >
	{
	  template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		     , typename Value
		     , typename ... Parameters
		   >
	  static auto apply
	  (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	   parameter::named_parameter<parameter::tag::probability,Value> parameter,
	   Parameters&& ... parameters)
	  {
	    return beta_distribution<tag::probability_precision>
	      (pack.add(parameter),std::forward<Parameters>(parameters)...);
	  }
	};

	template <>
	struct retag < parameter::tag::precision >
	{
	  template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		     , typename Value
		     , typename ... Parameters
		   >
	  static auto apply
	  (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	   parameter::named_parameter<parameter::tag::precision,Value> parameter,
	   Parameters&& ... parameters)
	  {
	    return beta_distribution<tag::probability_precision>
	      (pack.add(parameter),std::forward<Parameters>(parameters)...);
	  }
	};

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   >
	auto beta_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack)
	{ return make_beta_strategy<Tag>(pack); }

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Key, typename Value
		   , typename ... Parameters
		   >
	auto beta_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	 parameter::named_parameter<Key,Value> parameter,
	 Parameters&& ... parameters)
	{
	  return beta_distribution<Tag>
	    (pack.add(parameter),
	     std::forward<Parameters>(parameters)...);
	}

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Value
		   , typename ... Parameters
		   >
	auto beta_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	 Value&& value,
	 Parameters&& ... parameters)
	{
	  return beta_distribution<Tag>
	    (pack.add(std::forward<Value>(value)),
	     std::forward<Parameters>(parameters)...);
	}

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged >
	auto beta_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack)
	{ return beta_distribution<tag::alpha_beta>(pack); }

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Key, typename Value
		   , typename ... Parameters
		   >
	auto beta_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	 parameter::named_parameter<Key,Value> parameter,
	 Parameters&& ... parameters)
	{
	  return detail::retag<Key>::apply(pack,parameter,std::forward<Parameters>(parameters)...);
	}

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Value
		   , typename ... Parameters
		   >
	auto beta_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	 Value&& value,
	 Parameters&& ... parameters)
	{
	  return beta_distribution
	    (pack.add(std::forward<Value>(value)),
	     std::forward<Parameters>(parameters)...);
	}

      }	// namespace detail

      template < typename ... Parameters >
      auto beta_distribution (Parameters&& ... parameters)
      {
	using boost::hana::make_map;
	using parameter::make_parameter;
	using parameter::make_parameter_pack;
	using vertex::make_expression_function;
	return detail::beta_distribution
	  (make_parameter_pack
	   (make_parameter<parameter::tag::alpha>().default_value(1),
	     make_parameter<parameter::tag::beta>().default_value(1),
	     make_parameter<parameter::tag::variate>().default_value(0.5),
	     make_parameter<parameter::tag::probability>().default_value(0.5),
	    make_parameter<parameter::tag::precision>().default_value(2)),
	   std::forward<Parameters>(parameters)...);
      }

    } // namespace beta

  } // namespace distribution

} // namespace graph_model

#endif
