/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_TAGGED_PARAMETER_HPP
#define GRAPH_MODEL_VERTEX_TAGGED_PARAMETER_HPP 1

#include <tuple>
#include <graph_model/vertex/parameter_tags.hpp>
#include <graph_model/vertex/expression_function.hpp>
#include <mpl/enable_constructor.hpp>

#include <iosfwd>
#include <debugging/class_name.hpp>

namespace graph_model
{

  namespace vertex
  {

    /// \brief Decorate an expression function with a tag
    ///
    template < typename Tag, typename ExprFunction >
    class tagged_parameter
    {
      using self = tagged_parameter;
      using pair_type = std::pair<Tag,ExprFunction>;

    public:
      using category = Tag;

      tagged_parameter () = default;
      template < typename F
		 , typename = mpl::enable_constructor_t<self,F>
		 >
      explicit tagged_parameter (F&& f)
	: pair_(Tag(),std::forward<F>(f))
      {}

      auto operator () () const { return expr()(); }

      auto& expr () & { return pair_.second; }
      auto& expr () const& { return pair_.second; }
      auto&& expr () && { return std::move(pair_.second); }

      auto& pair () & { return pair_; }
      auto& pair () const& { return pair_; }
      auto&& pair () && { return std::move(pair_); }

      template < typename NewTag >
      auto retag () &&
      { return tagged_parameter<NewTag,ExprFunction>(std::move(expr())); }

    private:
      pair_type pair_;
    };

    template < typename Tag, typename ExprFunction >
    std::ostream& operator << (std::ostream& os, tagged_parameter<Tag,ExprFunction>)
    { return os << debugging::full_class_name::apply<tagged_parameter<Tag,ExprFunction>>(); }

    template < typename Tag, typename ExprFunction >
    auto make_tagged_parameter (tagged_parameter<Tag,ExprFunction> const& parameter)
    { return parameter; }

    template < typename Tag, typename ExprFunction >
    auto make_tagged_parameter (tagged_parameter<Tag,ExprFunction>&& parameter)
    { return std::move(parameter); }

    template < typename Tag, typename T >
    auto make_tagged_parameter (T&& t)
    {
      auto f = make_expression_function(std::forward<T>(t));
      using function_type = decltype(f);
      using tagged_parameter_type = tagged_parameter<Tag,function_type>;
      auto result = tagged_parameter_type(std::move(f));
      return result;
    }

    template < typename T >
    auto make_tagged_parameter (T&& t)
    { return make_tagged_parameter<tag::unknown>(std::forward<T>(t)); }

  } // namespace vertex

} // namespace graph_model

#endif
