/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_FACTOR_VERTEX_DESCRIPTOR_HPP
#define GRAPH_MODEL_VERTEX_FACTOR_VERTEX_DESCRIPTOR_HPP 1

#include <iosfwd>
#include <boost/operators.hpp>
#include <observer_ptr.hpp>

namespace graph_model
{

  namespace vertex
  {

    template < typename Vertex >
    class vertex_descriptor
      : boost::totally_ordered<vertex_descriptor<Vertex>>
    {
      using self = vertex_descriptor;
      using pointer = observer_ptr<Vertex const>;

      template < typename Address >
      struct id_type
      {
	explicit id_type (Address address) : address_(address) {}

	friend std::ostream& operator << (std::ostream& os, id_type id)
	{ return os << "(" << id.address_ << ")"; }

      private:
	Address address_;
      };

      template < typename Address >
      auto make_id (Address address) const
      { return id_type<Address>(address); }

      auto make_id () const { return make_id(p_.get()); }

      auto& vertex () { assert(p_); return *p_; }
      auto& vertex () const { assert(p_); return *p_; }

    public:
      vertex_descriptor () = default;
      vertex_descriptor (pointer p) : p_(p) {}

      vertex_descriptor (Vertex const& v)
	: vertex_descriptor(make_observer(v))
      {}

      bool is_null_vertex () const { return !p_; }
      auto id () const { return make_id(); }
      auto identifier () const { return vertex().identifier(); }
      auto vertex_category () const { return vertex().vertex_category(); }

      auto out_degree () const { return vertex().out_degree(); }
      auto out_edges () const { return vertex().out_edges(); }

      friend bool operator == (self left, self right) { return left.p_ == right.p_; }
      friend bool operator < (self left, self right) { return left.p_ < right.p_; }

      friend std::ostream& operator << (std::ostream& os, self u)
      { return os << "(" << u.identifier() << "," << u.vertex_category() << "," << u.p_.get() << ")"; }

    private:
      pointer p_;
    };

    template < typename Vertex >
    auto make_vertex_descriptor (Vertex const& vertex)
    { return vertex_descriptor<Vertex>(vertex); }

  } // namespace vertex

} // namespace graph_model

#endif
