/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef SAMPLE_SPACE_RANDOM_DIRICHLET_DISTRIBUTION_HPP
#define SAMPLE_SPACE_RANDOM_DIRICHLET_DISTRIBUTION_HPP 1

#include <boost/operators.hpp>
#include <boost/random/gamma_distribution.hpp>
#include <boost/range/combine.hpp>
#include <always_false.hpp>
#include <traits/value.hpp>
#include <traits/variate.hpp>
#include <type_traits/is_map.hpp>
#include <sample_space/math/divide_assign.hpp>
#include <sample_space/math/norm_1.hpp>

namespace sample_space
{
  
namespace random
{

  template < typename Variate
	     , typename Alpha = Variate
	     >
  class dirichlet_distribution
    : boost::equality_comparable< dirichlet_distribution<Variate,Alpha> >
  {
    using self = dirichlet_distribution;
    using variate_type = traits::variate_t<Variate>;
    using alpha_type = traits::variate_t<Alpha>;

    using value_type = traits::value_t<variate_type>;
    using gamma_distribution = boost::random::gamma_distribution<value_type>;
    
  public:
    using result_type = Variate;
    
  public:
    struct param_type
      : boost::equality_comparable<param_type>
    {
      param_type () = default;
      param_type (Alpha alpha) : alpha_(alpha) {}

      Alpha const& alpha () const { return alpha_; }

      friend bool operator == (param_type left, param_type right)
      { return left.param_ == right.param_; }
      
    private:
      Alpha alpha_;
    };

  public:
    dirichlet_distribution () = default;
    dirichlet_distribution (Alpha alpha) : param_(alpha) {}
    dirichlet_distribution (param_type p) : param_(p) {}

    param_type const& param () const { return param_; }
    void param (param_type p) { param_ = p; }
    
    void reset () { gamma_.reset(); }

    template < typename Engine >
    result_type operator () (Engine& engine)
    { return generate_variate(engine,param_); }
    
    template < typename Engine >
    result_type operator () (Engine& engine, param_type param)
    { return generate_variate(engine,param); }
    
    friend bool operator == (self left, self right)
    { return left.param_ == right.param_; }

  private:
    template < typename Engine >
    result_type generate_variate (Engine& engine, param_type param)
    { return generate_variate(engine,param,type_traits::is_map_t<Variate>(),type_traits::is_map_t<Alpha>()); }
    
    template < typename Engine >
    result_type generate_variate (Engine& engine, param_type param, std::false_type, std::false_type)
    {
      // XXX<param_type> xxx;
      using boost::get;
      using sample_space::math::divide_assign;
      using sample_space::math::norm_1;
      result_type result (param.alpha().size());;
      for (auto pair : boost::combine(result,param.alpha()))
	{
	  // XXX<decltype(pair)> xxx;
	  get<0>(pair) = gamma(engine,get<1>(pair));
	}
      divide_assign(result,norm_1(result));
      return result;
    }

    template < typename Engine >
    result_type generate_variate (Engine& engine, param_type param, std::false_type, std::true_type)
    {
      using boost::get;
      using sample_space::math::divide_assign;
      using sample_space::math::norm_1;
      result_type result (param.alpha().size());;
      for (auto pair : boost::combine(result,param.alpha()))
	get<0>(pair) = gamma(engine,get<1>(pair).second);
      divide_assign(result,norm_1(result));
      return result;
    }

    template < typename Engine >
    result_type generate_variate (Engine& engine, param_type param, std::true_type, std::false_type)
    {
      static_assert(always_false<Engine>::value,
		    "mismatch between Variate and Alpha; "
		    "if Variate has an index, so must Alpha");
    }
    
    template < typename Engine >
    result_type generate_variate (Engine& engine, param_type param, std::true_type, std::true_type)
    {
      using boost::get;
      using sample_space::math::divide_assign;
      using sample_space::math::norm_1;
      result_type result;
      for (auto pair : param.alpha())
	result[get<0>(pair)] = gamma(engine,get<1>(pair));
      divide_assign(result,norm_1(result));
      return result;
    }

    template < typename Engine, typename Value >
    Value gamma (Engine& engine, Value alpha)
    {
      using param_type = typename gamma_distribution::param_type;
      return gamma_(engine,param_type(alpha));
    }
    
  private:
    gamma_distribution gamma_;
    param_type param_;    
  };
  
} // namespace random

} // namespace sample_space

#endif
